/*
Glock - a simple analog clock using GTK+
$Id: glock.c 7 2005-02-19 19:14:12Z eric $
Copyright 2005 Eric L. Smith <eric@brouhaha.com>

Glock is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 as
published by the Free Software Foundation.  Note that I am not
granting permission to redistribute or modify glock under the
terms of any later version of the General Public License.

Glock is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with this program (in the file "COPYING"); if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111, USA.
*/

#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>


static GtkWidget *main_window;
static GtkWidget *drawing_area;
static GtkWidget *menubar;  /* actually a popup menu in transparency/shape mode */

GdkColor black;
GdkColor white;


#define QMAKESTR(x) #x
#define MAKESTR(x) QMAKESTR(x)


char *progname;  /* set to argv[0] in main */


void usage (FILE *f)
{
  fprintf (f, "Glock %s:  analog clock\n",
	   MAKESTR(GLOCK_RELEASE));
  fprintf (f, "Copyright 2005 Eric L. Smith\n");
  fprintf (f, "http://www.brouhaha.com/~eric/software/glock/\n");
  fprintf (f, "\n");
  fprintf (f, "usage: %s [options...]\n", progname);
  fprintf (f, "options:\n");
  fprintf (f, "   --size n\n");
}


/* generate fatal error message to stderr, doesn't return */

void fatal (int ret, char *format, ...)
{
  va_list ap;

  fprintf (stderr, "fatal error: ");
  va_start (ap, format);
  vfprintf (stderr, format, ap);
  va_end (ap);
  if (ret == 1)
    usage (stderr);
  exit (ret);
}


void *alloc (size_t size)
{
  void *p;

  p = calloc (1, size);
  if (! p)
    fatal (2, "Memory allocation failed\n");
  return (p);
}


char *newstr (char *orig)
{
  int len;
  char *r;

  len = strlen (orig);
  r = (char *) alloc (len + 1);
  memcpy (r, orig, len + 1);
  return (r);
}


static void quit_callback (GtkWidget *widget, gpointer data)
{
  gtk_main_quit ();
}


static void edit_copy (GtkWidget *widget, gpointer data)
{
  /* $$$ not yet implemented */
}


static void help_about (GtkWidget *widget, gpointer data)
{
  GtkWidget *dialog;
  char buf [200];

  dialog = gtk_dialog_new_with_buttons ("About glock",
					GTK_WINDOW (main_window),
					GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_OK,
					GTK_RESPONSE_NONE,
					NULL);

  gtk_dialog_set_has_separator (GTK_DIALOG (dialog), TRUE);

  sprintf (buf, "glock release %s", MAKESTR(GLOCK_RELEASE));
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox),
		     gtk_label_new (buf));
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox),
		     gtk_label_new ("Analog clock\n"
				    "Copyright 2005 Eric L. Smith\n"
				    "http://www.brouhaha.com/~eric/software/glock/"));
  gtk_widget_show_all (dialog);
  gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_destroy (dialog);
}


static GtkItemFactoryEntry menu_items [] =
  {
    { "/_File",         NULL,         NULL,          0, "<Branch>" },
    { "/File/_Quit",    "<CTRL>Q",    gtk_main_quit, 0, "<StockItem>", GTK_STOCK_QUIT },
    { "/_Edit",         NULL,         NULL,          0, "<Branch>" },
    { "/Edit/_Copy",    "<control>C", edit_copy,     0, "<StockItem>", GTK_STOCK_COPY },
    { "/_Help",         NULL,         NULL,          0, "<LastBranch>" },
    { "/_Help/About",   NULL,         help_about,    0, "<Item>" }
  };

static gint nmenu_items = sizeof (menu_items) / sizeof (GtkItemFactoryEntry);


static GtkWidget *create_menus (GtkWidget *window,
				GtkType container_type)
{
  GtkAccelGroup *accel_group;
  GtkItemFactory *item_factory;

  accel_group = gtk_accel_group_new ();
  item_factory = gtk_item_factory_new (container_type,
				       "<main>",
				       accel_group);
  gtk_item_factory_create_items (item_factory, nmenu_items, menu_items, NULL);
  gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);
  return (gtk_item_factory_get_widget (item_factory, "<main>"));
}


gboolean window_click_callback (GtkWidget *widget, GdkEventButton *event)
{
  if (event->type == GDK_BUTTON_PRESS)
    {
      switch (event->button)
	{
	case 1:  /* left button */
	  gtk_window_begin_move_drag (GTK_WINDOW (main_window),
				      event->button,
				      event->x_root,
				      event->y_root,
				      event->time);
	  break;
	case 3:  /* right button */
	  gtk_menu_popup (GTK_MENU (menubar),
			  NULL,  /* parent_menu_shell */
			  NULL,  /* parent_menu_item */
			  NULL,  /* func */
			  NULL,  /* data */
			  event->button,
			  event->time);
	  break;
	}
    }
  return FALSE;
}


gboolean clock_tick (gpointer data)
{
  GdkRectangle rect = { 0, 0, 0, 0 };

  rect.width = drawing_area->allocation.width;
  rect.height = drawing_area->allocation.height;

  gdk_window_invalidate_rect (drawing_area->window,
			      & rect,
			      FALSE);

  return TRUE;
}


gboolean expose_callback (GtkWidget *widget,
			  GdkEventExpose *event,
			  gpointer data)
{
  int width = widget->allocation.width;
  int height = widget->allocation.height;
  int i;
  time_t t;
  struct tm tm;
  double hour_angle, minute_angle, second_angle;

  t = time (NULL);
  localtime_r (& t, & tm);

  gdk_draw_arc (widget->window,
		widget->style->fg_gc [GTK_WIDGET_STATE (widget)],
		TRUE,
		0,
		0,
		width,
		height,
		0,
		64 * 360);
  gdk_draw_arc (widget->window,
		widget->style->bg_gc [GTK_WIDGET_STATE (widget)],
		TRUE,
		width / 16,
		height / 16,
		(width * 7) / 8,
		(height * 7) / 8,
		0,
		64 * 360);

  for (i = 0; i < 12; i++)
    {
      hour_angle = i * M_PI / 6;

      gdk_draw_line (widget->window,
		     widget->style->fg_gc [GTK_WIDGET_STATE (widget)],
		     width / 2 + (width * 3 / 8) * sin (hour_angle),
		     height / 2 - (height * 3 / 8) * cos (hour_angle),
		     width / 2 + (width / 2) * sin (hour_angle),
		     height / 2 - (height / 2) * cos (hour_angle));
    }

  hour_angle = ((tm.tm_hour % 12) + (tm.tm_min / 60.0)) * M_PI / 6;
  minute_angle = (tm.tm_min + (tm.tm_sec / 60.0)) * M_PI / 30;
  second_angle = tm.tm_sec * M_PI / 30;

  gdk_draw_line (widget->window,
		 widget->style->fg_gc [GTK_WIDGET_STATE (widget)],
		 width / 2,
		 height / 2,
		 width / 2 + (width / 4) * sin (hour_angle),
		 height / 2 - (height / 4) * cos (hour_angle));

  gdk_draw_line (widget->window,
		 widget->style->fg_gc [GTK_WIDGET_STATE (widget)],
		 width / 2,
		 height / 2,
		 width / 2 + (width * 5 / 16) * sin (minute_angle),
		 height / 2 - (height * 5 / 16) * cos (minute_angle));

  gdk_draw_line (widget->window,
		 widget->style->fg_gc [GTK_WIDGET_STATE (widget)],
		 width / 2,
		 height / 2,
		 width / 2 + (width / 2) * sin (second_angle),
		 height / 2 - (height / 2) * cos (second_angle));

  return TRUE;
}


int main (int argc, char *argv[])
{
  int width = 200;
  int height = 200;

  GdkColormap *colormap;
  GdkGC *gc;
  GdkBitmap *window_shape_bitmap;
  GtkWidget *event_box;

  progname = newstr (argv [0]);

  gtk_init (& argc, & argv);

  while (--argc)
    {
      argv++;
      if (*argv [0] == '-')
	{
	  if (strcasecmp (argv [0], "--size") == 0)
	    {
	      if ((! --argc) || (! isdigit (*argv [1])))
		fatal (1, "size option requires numeric argument\n");
	      width = height = atoi (argv [1]);
	      argv++;
	    }
	  else if (strcasecmp (argv [0], "--help") == 0)
	    {
	      usage (stderr);
	      exit (0);
	    }
	  else
	    fatal (1, "unrecognized option '%s'\n", argv [0]);
	}
      else
	fatal (1, "unrecognized option '%s'\n", argv [0]);
    }

  main_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

  colormap = gdk_colormap_get_system ();

  gdk_color_black (colormap, & black);
  gdk_color_white (colormap, & white);

  window_shape_bitmap = (GdkBitmap *) gdk_pixmap_new (NULL,
						      width,
						      height,
						      1);

  gc = gdk_gc_new (window_shape_bitmap);

  gdk_gc_set_foreground (gc, & black);
  gdk_gc_set_background (gc, & white);

  // fill window_shape_bitmap with black

  gdk_draw_rectangle (window_shape_bitmap,
		      gc,
		      TRUE,  // filled
		      0,     // x
		      0,     // y
		      width,
		      height);

  gdk_gc_set_foreground (gc, & white);
  gdk_gc_set_background (gc, & black);

  // draw white filled circle into window_shape_bitmap
  gdk_draw_arc (window_shape_bitmap,
		gc,
		TRUE,     // filled
		0,        // x
		0,        // y
		width,
		height,
		0,        // angle 1
		360*64);  // angle 2: full circle

  gtk_window_set_resizable (GTK_WINDOW (main_window), FALSE);

  gtk_window_set_title (GTK_WINDOW (main_window),
			progname);

  menubar = create_menus (main_window, GTK_TYPE_MENU);

  gtk_widget_shape_combine_mask (main_window,
				 window_shape_bitmap,
				 0,
				 0);

  drawing_area = gtk_drawing_area_new ();

  gtk_widget_set_size_request (drawing_area, width, height);

  g_signal_connect (G_OBJECT (drawing_area),
		    "expose_event",
		    G_CALLBACK (expose_callback),
		    NULL);

  event_box = gtk_event_box_new ();
  gtk_container_add (GTK_CONTAINER (event_box), drawing_area);

  gtk_container_add (GTK_CONTAINER (main_window), event_box);

  gtk_window_set_decorated (GTK_WINDOW (main_window), FALSE);

  // Have to show everything again, now that we've done display_init()
  // and combined the shape mask.
  gtk_widget_show_all (main_window);

  // Mouse click to drag
  g_signal_connect (G_OBJECT (main_window),
		    "button_press_event",
		    G_CALLBACK (window_click_callback),
		    NULL);

  g_signal_connect (G_OBJECT (main_window),
		    "destroy",
		    G_CALLBACK (quit_callback),
		    NULL);

  g_timeout_add (1000, clock_tick, NULL);

  gtk_main ();

  exit (0);
}
